#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        ["data/_test/aggregate/csms.txt", "data/_test/aggregate/xls.txt"],
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    assert len(pr["crosslinks"]) == 10
    unique_peptide = unique(pr, by="peptide")
    assert len(unique_peptide["crosslink-spectrum-matches"]) == 5
    assert len(unique_peptide["crosslinks"]) == 3


def test2():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        ["data/_test/aggregate/csms.txt", "data/_test/aggregate/xls.txt"],
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    assert len(pr["crosslinks"]) == 10
    unique_protein = unique(pr, by="protein")
    assert len(unique_protein["crosslink-spectrum-matches"]) == 5
    assert len(unique_protein["crosslinks"]) == 2


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_peptide = aggregate(pr["crosslink-spectrum-matches"], by="peptide")
    assert len(aggregate_peptide) == 3


def test4():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_protein = aggregate(pr["crosslink-spectrum-matches"], by="protein")
    assert len(aggregate_protein) == 2


def test5():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/csms.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    u = unique(pr["crosslink-spectrum-matches"])
    assert len(u) == 5
    assert [csm["alpha_peptide"] for csm in u] == [
        "KPEPTIDE",
        "KPEPTIDE",
        "PEKPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [csm["alpha_proteins"] for csm in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]


def test6():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"])
    assert len(u) == 3
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEKPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]


def test7():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"], by="protein")
    assert len(u) == 2
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [["PROTA"], ["PROTA", "PROTB"]]


def test8():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_peptide = aggregate(pr["crosslink-spectrum-matches"], by="peptide")
    assert len(aggregate_peptide) == 3
    assert [xl["alpha_peptide"] for xl in aggregate_peptide] == [
        "KPEPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [xl["alpha_proteins"] for xl in aggregate_peptide] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]


def test9():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read("data/_test/aggregate/csms.txt", engine="custom", crosslinker="DSS")
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_protein = aggregate(pr["crosslink-spectrum-matches"], by="protein")
    assert len(aggregate_protein) == 2
    assert [xl["alpha_peptide"] for xl in aggregate_protein] == ["KPEPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in aggregate_protein] == [
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]


def test10():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_min.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"])
    assert len(u) == 3
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEKPTIDE", "PEPKTIDE"]


def test11():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_min.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    err_str = (
        r"Grouping by protein crosslink position is only available if all crosslinks have defined protein crosslink positions!\n"
        r"This error might be fixable with 'transform\.reannotate_positions\(\)'\!"
    )
    with pytest.raises(ValueError, match=err_str):
        _u = unique(pr["crosslinks"], by="protein")


def test12():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_min.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_peptide = aggregate(pr["crosslink-spectrum-matches"])
    assert len(aggregate_peptide) == 3
    assert [xl["alpha_peptide"] for xl in aggregate_peptide] == [
        "KPEPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]


def test13():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_min.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Grouping by protein crosslink position is only available if all crosslink-spectrum-matches have defined protein crosslink positions!\n"
        r"This error might be fixable with 'transform\.reannotate_positions\(\)'\!"
    )
    with pytest.raises(ValueError, match=err_str):
        _aggregate_protein = aggregate(pr["crosslink-spectrum-matches"], by="protein")


def test14():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    u = unique(pr["crosslink-spectrum-matches"])
    assert len(u) == 5
    assert [csm["alpha_peptide"] for csm in u] == [
        "KPEPTIDE",
        "KPEPTIDE",
        "PEKPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [csm["alpha_proteins"] for csm in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTB", "PROTA"],
    ]
    assert u[0]["score"] == pytest.approx(2.5)
    assert u[1]["score"] == pytest.approx(4.5)
    assert u[2]["score"] == pytest.approx(6.5)
    assert u[3]["score"] == pytest.approx(8.5)
    assert u[4]["score"] == pytest.approx(10.5)


def test15():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    u = unique(pr["crosslink-spectrum-matches"], score="lower_better")
    assert len(u) == 5
    assert [csm["alpha_peptide"] for csm in u] == [
        "KPEPTIDE",
        "KPEPTIDE",
        "PEKPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [csm["alpha_proteins"] for csm in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]
    assert u[0]["score"] == pytest.approx(1.5)
    assert u[1]["score"] == pytest.approx(3.5)
    assert u[2]["score"] == pytest.approx(5.5)
    assert u[3]["score"] == pytest.approx(7.5)
    assert u[4]["score"] == pytest.approx(9.5)


def test16():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"])
    assert len(u) == 3
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEKPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTB", "PROTA"],
    ]
    assert u[0]["score"] == pytest.approx(4.5)
    assert u[1]["score"] == pytest.approx(8.5)
    assert u[2]["score"] == pytest.approx(10.5)


def test17():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"], score="lower_better")
    assert len(u) == 3
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEKPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]
    assert u[0]["score"] == pytest.approx(1.5)
    assert u[1]["score"] == pytest.approx(5.5)
    assert u[2]["score"] == pytest.approx(9.5)


def test18():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"], by="protein")
    assert len(u) == 2
    assert [xl["alpha_peptide"] for xl in u] == ["PEKPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [["PROTA"], ["PROTB", "PROTA"]]
    assert u[0]["score"] == pytest.approx(8.5)
    assert u[1]["score"] == pytest.approx(10.5)


def test19():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/xls_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslinks"]) == 10
    u = unique(pr["crosslinks"], by="protein", score="lower_better")
    assert len(u) == 2
    assert [xl["alpha_peptide"] for xl in u] == ["KPEPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in u] == [["PROTA"], ["PROTA", "PROTB"]]
    assert u[0]["score"] == pytest.approx(1.5)
    assert u[1]["score"] == pytest.approx(9.5)


def test20():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_peptide = aggregate(pr["crosslink-spectrum-matches"])
    assert len(aggregate_peptide) == 3
    assert [xl["alpha_peptide"] for xl in aggregate_peptide] == [
        "KPEPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [xl["alpha_proteins"] for xl in aggregate_peptide] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTB", "PROTA"],
    ]
    assert aggregate_peptide[0]["score"] == pytest.approx(4.5)
    assert aggregate_peptide[1]["score"] == pytest.approx(8.5)
    assert aggregate_peptide[2]["score"] == pytest.approx(10.5)


def test21():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_peptide = aggregate(
        pr["crosslink-spectrum-matches"], score="lower_better"
    )
    assert len(aggregate_peptide) == 3
    assert [xl["alpha_peptide"] for xl in aggregate_peptide] == [
        "KPEPTIDE",
        "PEKPTIDE",
        "PEPKTIDE",
    ]
    assert [xl["alpha_proteins"] for xl in aggregate_peptide] == [
        ["PROTA"],
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]
    assert aggregate_peptide[0]["score"] == pytest.approx(1.5)
    assert aggregate_peptide[1]["score"] == pytest.approx(5.5)
    assert aggregate_peptide[2]["score"] == pytest.approx(9.5)


def test22():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_protein = aggregate(pr["crosslink-spectrum-matches"], by="protein")
    assert len(aggregate_protein) == 2
    assert [xl["alpha_peptide"] for xl in aggregate_protein] == ["PEKPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in aggregate_protein] == [
        ["PROTA"],
        ["PROTB", "PROTA"],
    ]
    assert aggregate_protein[0]["score"] == pytest.approx(8.5)
    assert aggregate_protein[1]["score"] == pytest.approx(10.5)


def test23():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    aggregate_protein = aggregate(
        pr["crosslink-spectrum-matches"], by="protein", score="lower_better"
    )
    assert len(aggregate_protein) == 2
    assert [xl["alpha_peptide"] for xl in aggregate_protein] == ["KPEPTIDE", "PEPKTIDE"]
    assert [xl["alpha_proteins"] for xl in aggregate_protein] == [
        ["PROTA"],
        ["PROTA", "PROTB"],
    ]
    assert aggregate_protein[0]["score"] == pytest.approx(1.5)
    assert aggregate_protein[1]["score"] == pytest.approx(9.5)


def test24():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
        r"peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
    )
    with pytest.raises(TypeError, match=err_str):
        _u = unique(pr["crosslink-spectrum-matches"], by="sequence")


def test25():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
        r"peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
    )
    with pytest.raises(TypeError, match=err_str):
        _a = aggregate(pr["crosslink-spectrum-matches"], by="sequence")


def test26():
    from pyXLMS.parser import read
    from pyXLMS.transform import unique

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
        r"-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
    )
    with pytest.raises(TypeError, match=err_str):
        _u = unique(pr["crosslink-spectrum-matches"], score="lower")


def test27():
    from pyXLMS.parser import read
    from pyXLMS.transform import aggregate

    pr = read(
        "data/_test/aggregate/csms_score.txt",
        engine="custom",
        crosslinker="DSS",
    )
    assert len(pr["crosslink-spectrum-matches"]) == 10
    err_str = (
        r"Parameter 'score' has to be one of 'higher_better' or 'lower_better'! If two identical crosslinks or crosslink-spectrum"
        r"-matches are found, the one with the higher score is kept if 'higher_better' is selected, and vice versa."
    )
    with pytest.raises(TypeError, match=err_str):
        _a = aggregate(pr["crosslink-spectrum-matches"], score="lower")



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest

XIFDR_CSMS = "data/xi/1perc_xl_boost_CSM_xiFDR2.2.1.csv"
XIFDR_LINKS = "data/xi/1perc_xl_boost_Links_xiFDR2.2.1.csv"


def test1():
    from pyXLMS.exporter import get_msannika_crosslink_sequence

    assert get_msannika_crosslink_sequence("PEPKTIDE", 4) == "PEP[K]TIDE"
    assert get_msannika_crosslink_sequence("KPEPTIDE", 1) == "[K]PEPTIDE"
    assert get_msannika_crosslink_sequence("PEPTIDEK", 8) == "PEPTIDE[K]"
    with pytest.raises(
        ValueError,
        match=r"Crosslink position outside of range! Must be in range \[1, 8\]\.",
    ):
        _seq = get_msannika_crosslink_sequence("PEPTIDEK", 9)


def test2():
    from pyXLMS.exporter import to_msannika
    from pyXLMS.data import create_crosslink_min

    with pytest.raises(
        TypeError,
        match="Parameter 'format' has to be one of 'csv', 'tsv', or 'xlsx'!",
    ):
        xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
        xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
        crosslinks = [xl1, xl2]
        _df = to_msannika(crosslinks, "crosslinks.parquet", format="parquet")


def test3():
    from pyXLMS.exporter import to_msannika

    with pytest.raises(
        ValueError,
        match="Provided data does not contain any crosslinks or crosslink-spectrum-matches!",
    ):
        _df = to_msannika([])


def test4():
    from pyXLMS.exporter import to_msannika
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]
    df = to_msannika(crosslinks)
    assert df.shape[0] == 2
    assert df.shape[1] == 11


def test5():
    from pyXLMS.exporter import to_msannika
    from pyXLMS.data import create_crosslink_min

    xl1 = create_crosslink_min("KPEPTIDE", 1, "PKEPTIDE", 2)
    xl2 = create_crosslink_min("PEKPTIDE", 3, "PEPKTIDE", 4)
    crosslinks = [xl1, xl2]
    df = to_msannika(crosslinks, filename="crosslinks.csv", format="csv")
    assert df.shape[0] == 2
    assert df.shape[1] == 11


def test6():
    from pyXLMS.exporter import to_msannika
    from pyXLMS.data import create_csm_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    csm2 = create_csm_min("PEKPTIDE", 3, "PEPKTIDE", 4, "RUN_1", 2)
    csms = [csm1, csm2]
    df = to_msannika(csms)
    assert df.shape[0] == 2
    assert df.shape[1] == 20


def test7():
    from pyXLMS.exporter import to_msannika
    from pyXLMS.data import create_csm_min

    csm1 = create_csm_min("KPEPTIDE", 1, "PKEPTIDE", 2, "RUN_1", 1)
    csm2 = create_csm_min("PEKPTIDE", 3, "PEPKTIDE", 4, "RUN_1", 2)
    csms = [csm1, csm2]
    df = to_msannika(csms, filename="csms.csv", format="csv")
    assert df.shape[0] == 2
    assert df.shape[1] == 20


def test8():
    from pyXLMS import parser as p
    from pyXLMS import exporter as e

    xi = p.read_xi(XIFDR_CSMS, parse_modifications=False, modifications={}, verbose=0)
    _ = e.to_msannika(
        xi["crosslink-spectrum-matches"], filename="xi_csms_to_annika.csv", format="csv"
    )
    with pytest.raises(KeyError, match="Modifications A"):
        _pr = p.read_msannika("xi_csms_to_annika.csv", sep=",")


def test9():
    from pyXLMS import parser as p
    from pyXLMS import exporter as e
    from pyXLMS.transform import modifications_to_str as mts

    xi = p.read_xi(XIFDR_CSMS, verbose=0)
    _ = e.to_msannika(
        xi["crosslink-spectrum-matches"], filename="xi_csms_to_annika.csv", format="csv"
    )
    pr = p.read_msannika("xi_csms_to_annika.csv", sep=",", parse_modifications=False)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "MS Annika"
    assert pr["crosslink-spectrum-matches"] is not None
    assert pr["crosslinks"] is None

    csms = pr["crosslink-spectrum-matches"]
    assert len(csms) == 413

    csm = csms[0]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 1
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [575]
    assert csm["alpha_proteins_peptide_positions"] == [575]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "KIECFDSVEISGVEDR"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 1
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [575]
    assert csm["beta_proteins_peptide_positions"] == [575]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(27.268)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 19140
    assert csm["charge"] == 4
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[332]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "NPIDFLEAKGYK"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 9
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [1189]
    assert csm["alpha_proteins_peptide_positions"] == [1181]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QWYKNK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 4
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [488]
    assert csm["beta_proteins_peptide_positions"] == [485]
    assert csm["beta_score"] is None
    assert csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(10.071)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 21757
    assert csm["charge"] == 3
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None

    csm = csms[-1]
    assert csm["data_type"] == "crosslink-spectrum-match"
    assert csm["completeness"] == "partial"
    assert csm["alpha_peptide"] == "QITKHVAQILDSR"
    assert mts(csm["alpha_modifications"]) is None
    assert csm["alpha_peptide_crosslink_position"] == 4
    assert csm["alpha_proteins"] == ["Cas9"]
    assert csm["alpha_proteins_crosslink_positions"] == [933]
    assert csm["alpha_proteins_peptide_positions"] == [930]
    assert csm["alpha_score"] is None
    assert not csm["alpha_decoy"]
    assert csm["beta_peptide"] == "QQLPEKYK"
    assert mts(csm["beta_modifications"]) is None
    assert csm["beta_peptide_crosslink_position"] == 6
    assert csm["beta_proteins"] == ["Cas9"]
    assert csm["beta_proteins_crosslink_positions"] == [350]
    assert csm["beta_proteins_peptide_positions"] == [345]
    assert csm["beta_score"] is None
    assert not csm["beta_decoy"]
    assert csm["crosslink_type"] == "intra"
    assert csm["score"] == pytest.approx(6.808)
    assert csm["spectrum_file"] == "XLpeplib_Beveridge_QEx-HFX_DSS_R1.mgf"
    assert csm["scan_nr"] == 12190
    assert csm["charge"] == 5
    assert csm["retention_time"] is None
    assert csm["ion_mobility"] is None


def test10():
    from pyXLMS import parser as p
    from pyXLMS import exporter as e

    xi = p.read_xi(XIFDR_LINKS, parse_modifications=False, modifications={}, verbose=0)
    _ = e.to_msannika(xi["crosslinks"], filename="xi_xls_to_annika.xlsx", format="xlsx")
    pr = p.read_msannika("xi_xls_to_annika.xlsx", parse_modifications=False)
    assert pr["data_type"] == "parser_result"
    assert pr["completeness"] == "partial"
    assert pr["search_engine"] == "MS Annika"
    assert pr["crosslink-spectrum-matches"] is None
    assert pr["crosslinks"] is not None

    xls = pr["crosslinks"]
    assert len(xls) == 227

    xl = xls[0]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "VVDELVKVMGR"
    assert xl["alpha_peptide_crosslink_position"] == 7
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [753]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "VVDELVKVMGR"
    assert xl["beta_peptide_crosslink_position"] == 7
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [753]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(40.679)

    xl = xls[30]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "FDNLTKAER"
    assert xl["alpha_peptide_crosslink_position"] == 6
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [952]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "YDENDKLIR"
    assert xl["beta_peptide_crosslink_position"] == 6
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [906]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(25.616)

    xl = xls[219]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "NPIDFLEAKGYK"
    assert xl["alpha_peptide_crosslink_position"] == 9
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [488]
    assert xl["alpha_decoy"]
    assert xl["beta_peptide"] == "QWYKNK"
    assert xl["beta_peptide_crosslink_position"] == 4
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [1189]
    assert xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(10.071)

    xl = xls[-1]
    assert xl["data_type"] == "crosslink"
    assert xl["completeness"] == "full"
    assert xl["alpha_peptide"] == "LSKSR"
    assert xl["alpha_peptide_crosslink_position"] == 3
    assert xl["alpha_proteins"] == ["Cas9"]
    assert xl["alpha_proteins_crosslink_positions"] == [222]
    assert not xl["alpha_decoy"]
    assert xl["beta_peptide"] == "LSKSR"
    assert xl["beta_peptide_crosslink_position"] == 3
    assert xl["beta_proteins"] == ["Cas9"]
    assert xl["beta_proteins_crosslink_positions"] == [222]
    assert not xl["beta_decoy"]
    assert xl["crosslink_type"] == "intra"
    assert xl["score"] == pytest.approx(9.619)



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


@pytest.fixture(autouse=True)
def cleanup_figures():
    from matplotlib import use
    import matplotlib.pyplot as plt

    use("agg")

    yield

    plt.close(fig="all")


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test1():
    from pyXLMS import parser
    from pyXLMS import plotting

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    fig, ax = plotting.plot_venn_diagram(
        a, b, labels=["MS Annika", "MaxQuant"], colors=["orange", "blue"]
    )
    assert fig is not None
    assert ax is not None


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test2():
    from pyXLMS import parser
    from pyXLMS import plotting

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    c = parser.read(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
        engine="pLink",
        crosslinker="DSS",
    )
    c = c["crosslink-spectrum-matches"]
    fig, ax = plotting.plot_venn_diagram(
        a, b, c, labels=["MS Annika", "MaxQuant", "pLink"], contour=True
    )
    assert fig is not None
    assert ax is not None


def test3():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]

    err_str = r"Parameter 'by' has to be one of 'peptide' or 'protein'! Option 'peptide' will group by peptide sequence and "
    err_str += r"peptide crosslink position while option 'protein' will group by protein identifier and protein crosslink position."
    with pytest.raises(TypeError, match=err_str):
        _plot = plot_venn_diagram(a, b, by="sequence")


def test4():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]

    with pytest.raises(
        ValueError,
        match=r"Can't plot venn diagram if no crosslink-spectrum-matches or crosslinks are given in data_1!",
    ):
        _plot = plot_venn_diagram([], a)


def test5():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]

    with pytest.raises(
        ValueError,
        match=r"Can't plot venn diagram if no crosslink-spectrum-matches or crosslinks are given in data_2!",
    ):
        _plot = plot_venn_diagram(a, [])


def test6():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]

    with pytest.raises(
        ValueError,
        match=r"Can't plot 3-set venn diagram if no crosslink-spectrum-matches or crosslinks are given in data_3!",
    ):
        _plot = plot_venn_diagram(a, b, [])


def test7():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    a[0]["data_type"] = "peptide-spectrum-match"

    with pytest.raises(
        TypeError,
        match=r"Unsupported data type for input data! Parameter data_1 has to be a list of crosslink or crosslink-spectrum-match!",
    ):
        _plot = plot_venn_diagram(a, b)


def test8():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    a[0]["data_type"] = "peptide-spectrum-match"

    with pytest.raises(
        TypeError,
        match=r"Unsupported data type for input data! Parameter data_2 has to be a list of crosslink or crosslink-spectrum-match!",
    ):
        _plot = plot_venn_diagram(b, a)


def test9():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    c = parser.read(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
        engine="pLink",
        crosslinker="DSS",
    )
    c = c["crosslink-spectrum-matches"]
    a[0]["data_type"] = "peptide-spectrum-match"

    with pytest.raises(
        TypeError,
        match=r"Unsupported data type for input data! Parameter data_3 has to be a list of crosslink or crosslink-spectrum-match, or None!",
    ):
        _plot = plot_venn_diagram(c, b, a)


def test10():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    c = parser.read(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
        engine="pLink",
        crosslinker="DSS",
    )
    c = c["crosslink-spectrum-matches"]
    a[0]["completeness"] = "partial"
    a[0]["alpha_proteins"] = None

    err_str = r"Grouping by protein crosslink position is only available if all data have defined protein crosslink positions!\n"
    err_str += (
        r"This error might be fixable with 'transform\.reannotate_positions\(\)'!"
    )
    with pytest.raises(
        ValueError,
        match=err_str,
    ):
        _plot = plot_venn_diagram(c, b, a, by="protein")


@pytest.mark.filterwarnings("ignore:'mode' parameter is deprecated")
def test11():
    from pyXLMS import parser
    from pyXLMS.plotting import plot_venn_diagram

    a = parser.read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.txt",
        engine="MS Annika",
        crosslinker="DSS",
    )
    a = a["crosslink-spectrum-matches"]
    b = parser.read(
        "data/maxquant/run1/crosslinkMsms.txt", engine="MaxQuant", crosslinker="DSS"
    )
    b = b["crosslink-spectrum-matches"]
    c = parser.read(
        "data/plink2/Cas9_plus10_2024.06.20.filtered_cross-linked_spectra.csv",
        engine="pLink",
        crosslinker="DSS",
    )
    c = c["crosslink-spectrum-matches"]

    plot = plot_venn_diagram(c, b, a, by="protein")
    assert plot is not None



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import pandas as pd

from ..data import check_input
from ..data import check_input_multi

from typing import Optional
from typing import Dict
from typing import List
from typing import Tuple
from typing import Any


def __get_modified_peptide(
    sequence: str,
    modifications: Optional[Dict[int, Tuple[str, float]]],
    crosslink_position: int,
    crosslinker: Optional[str | float],
) -> str:
    r"""Returns the Proforma string for a single peptide.

    Parameters
    ----------
    sequence : str
        The unmodified peptide sequence.
    modifications : dict of int, tuple of str and float
        The pyXLMS specific modifications object. See ``data.create_csm()`` for reference.
    crosslink_position : int
        Crosslink position in the peptide sequence (1-based).
    crosslinker : str, or float, or None
        Optional name or mass of the crosslink reagent. If the name is given, it should be a valid
        name from XLMOD.

    Returns
    -------
    str
        The Proforma string of the peptidoform.

    Notes
    -----
    - This function should not be called directly, it is called from ``__to_proforma_csm()`` and ``__to_proforma_xl``.
    - Modifications with unknown mass are skipped.
    - If no modifications are given, only the crosslink modification will be encoded in the Proforma.
    - If no modifications are given and no crosslinker is given, the unmodified peptide Proforma will be returned.
    """
    if isinstance(crosslinker, float):
        crosslinker = f"+{crosslinker}" if crosslinker > 0.0 else f"{crosslinker}"
    pep_len = len(sequence)
    if modifications is not None:
        new_modifications = dict()
        for pos, mod in modifications.items():
            if not pd.isna(mod[1]):
                new_modifications[pos] = (
                    mod[0],
                    f"+{mod[1]}" if mod[1] > 0.0 else f"{mod[1]}",
                )
        if crosslink_position not in new_modifications and crosslinker is not None:
            new_modifications[crosslink_position] = ("", f"{crosslinker}")
        for pos in sorted(new_modifications.keys(), reverse=True):
            if pos == 0:
                sequence = f"[{new_modifications[pos][1]}]-" + sequence
            elif pos == pep_len + 1:
                sequence = sequence + f"-[{new_modifications[pos][1]}]"
            else:
                sequence = (
                    sequence[:pos] + f"[{new_modifications[pos][1]}]" + sequence[pos:]
                )
        return sequence
    if crosslinker is not None:
        sequence = (
            sequence[:crosslink_position]
            + f"[{crosslinker}]"
            + sequence[crosslink_position:]
        )
        return sequence
    return sequence


def __to_proforma_csm(csm: Dict[str, Any], crosslinker: Optional[str | float]) -> str:
    r"""Returns the Proforma string for a single crosslink-spectrum-match.

    Parameters
    ----------
    csm : dict of str, any
        A pyXLMS crosslink-spectrum-match object. See ``data.create_csm()``.
    crosslinker : str, or float, or None
        Optional name or mass of the crosslink reagent. If the name is given, it should be a valid
        name from XLMOD.

    Returns
    -------
    str
        The Proforma string of the crosslink-spectrum-match.

    Notes
    -----
    - This function should not be called directly, it is called from ``to_proforma()``.
    - Modifications with unknown mass are skipped.
    - If no modifications are given, only the crosslink modification will be encoded in the Proforma.
    - If no modifications are given and no crosslinker is given, the unmodified peptide Proforma will be returned.
    """
    peptide_a = __get_modified_peptide(
        csm["alpha_peptide"],
        csm["alpha_modifications"],
        csm["alpha_peptide_crosslink_position"],
        crosslinker,
    )
    peptide_b = __get_modified_peptide(
        csm["beta_peptide"],
        csm["beta_modifications"],
        csm["beta_peptide_crosslink_position"],
        crosslinker,
    )
    if csm["charge"] is not None:
        return f"{peptide_a}//{peptide_b}/{csm['charge']}"
    return f"{peptide_a}//{peptide_b}"


def __to_proforma_xl(xl: Dict[str, Any], crosslinker: Optional[str | float]) -> str:
    r"""Returns the Proforma string for a single crosslink.

    Parameters
    ----------
    xl : dict of str, any
        A pyXLMS crosslink object. See ``data.create_crosslink()``.
    crosslinker : str, or float, or None
        Optional name or mass of the crosslink reagent. If the name is given, it should be a valid
        name from XLMOD.

    Returns
    -------
    str
        The Proforma string of the crosslink.

    Notes
    -----
    - This function should not be called directly, it is called from ``to_proforma()``.
    - Modifications with unknown mass are skipped.
    - If no modifications are given, only the crosslink modification will be encoded in the Proforma.
    - If no modifications are given and no crosslinker is given, the unmodified peptide Proforma will be returned.
    """
    peptide_a = __get_modified_peptide(
        xl["alpha_peptide"], None, xl["alpha_peptide_crosslink_position"], crosslinker
    )
    peptide_b = __get_modified_peptide(
        xl["beta_peptide"], None, xl["beta_peptide_crosslink_position"], crosslinker
    )
    return f"{peptide_a}//{peptide_b}"


def to_proforma(
    data: Dict[str, Any] | List[Dict[str, Any]],
    crosslinker: Optional[str | float] = None,
) -> str | List[str]:
    r"""Returns the Proforma string for a single crosslink or crosslink-spectrum-match, or for
    a list of crosslinks or crosslink-spectrum-matches.

    Parameters
    ----------
    data : dict of str, any, or list of dict of str, any
        A pyXLMS crosslink object, e.g. see ``data.create_crosslink()``. Or a pyXLMS crosslink-spectrum-match
        object, e.g. see ``data.create_csm()``. Alternatively, a list of crosslinks or crosslink-spectrum-matches
        can also be provided.
    crosslinker : str, or float, or None, default = None
        Optional name or mass of the crosslink reagent. If the name is given, it should be a valid
        name from XLMOD. If the crosslink modification is contained in the crosslink-spectrum-match object
        this parameter has no effect.

    Returns
    -------
    str
        The Proforma string of the crosslink or crosslink-spectrum-match. If a list was provided
        a list containing all Proforma strings is returned.

    Raises
    ------
    TypeError
        If an unsupported data type is provided.

    Notes
    -----
    - Modifications with unknown mass are skipped.
    - If no modifications are given, only the crosslink modification will be encoded in the Proforma.
    - If no modifications are given and no crosslinker is given, the unmodified peptide Proforma will be returned.

    Examples
    --------
    >>> from pyXLMS.data import create_crosslink_min
    >>> from pyXLMS.transform import to_proforma
    >>> xl = create_crosslink_min("PEPKTIDE", 4, "KPEPTIDE", 1)
    >>> to_proforma(xl)
    'KPEPTIDE//PEPKTIDE'

    >>> from pyXLMS.data import create_crosslink_min
    >>> from pyXLMS.transform import to_proforma
    >>> xl = create_crosslink_min("PEPKTIDE", 4, "KPEPTIDE", 1)
    >>> to_proforma(xl, crosslinker="Xlink:DSSO")
    'K[Xlink:DSSO]PEPTIDE//PEPK[Xlink:DSSO]TIDE'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min("PEPKTIDE", 4, "KPEPTIDE", 1, "RUN_1", 1)
    >>> to_proforma(csm)
    'KPEPTIDE//PEPKTIDE'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min("PEPKTIDE", 4, "KPEPTIDE", 1, "RUN_1", 1)
    >>> to_proforma(csm, crosslinker="Xlink:DSSO")
    'K[Xlink:DSSO]PEPTIDE//PEPK[Xlink:DSSO]TIDE'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min(
    ...     "PEPKTIDE",
    ...     4,
    ...     "KPMEPTIDE",
    ...     1,
    ...     "RUN_1",
    ...     1,
    ...     modifications_b={3: ("Oxidation", 15.994915)},
    ... )
    >>> to_proforma(csm, crosslinker="Xlink:DSSO")
    'K[Xlink:DSSO]PM[+15.994915]EPTIDE//PEPK[Xlink:DSSO]TIDE'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min(
    ...     "PEPKTIDE",
    ...     4,
    ...     "KPMEPTIDE",
    ...     1,
    ...     "RUN_1",
    ...     1,
    ...     modifications_b={3: ("Oxidation", 15.994915)},
    ...     charge=3,
    ... )
    >>> to_proforma(csm, crosslinker="Xlink:DSSO")
    'K[Xlink:DSSO]PM[+15.994915]EPTIDE//PEPK[Xlink:DSSO]TIDE/3'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min(
    ...     "PEPKTIDE",
    ...     4,
    ...     "KPMEPTIDE",
    ...     1,
    ...     "RUN_1",
    ...     1,
    ...     modifications_a={4: ("DSSO", 158.00376)},
    ...     modifications_b={1: ("DSSO", 158.00376), 3: ("Oxidation", 15.994915)},
    ...     charge=3,
    ... )
    >>> to_proforma(csm)
    'K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3'

    >>> from pyXLMS.data import create_csm_min
    >>> from pyXLMS.transform import to_proforma
    >>> csm = create_csm_min(
    ...     "PEPKTIDE",
    ...     4,
    ...     "KPMEPTIDE",
    ...     1,
    ...     "RUN_1",
    ...     1,
    ...     modifications_a={4: ("DSSO", 158.00376)},
    ...     modifications_b={1: ("DSSO", 158.00376), 3: ("Oxidation", 15.994915)},
    ...     charge=3,
    ... )
    >>> to_proforma(csm, crosslinker="Xlink:DSSO")
    'K[+158.00376]PM[+15.994915]EPTIDE//PEPK[+158.00376]TIDE/3'
    """
    _ok = (
        check_input_multi(crosslinker, "crosslinker", [str, float])
        if crosslinker is not None
        else True
    )
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        return [to_proforma(item, crosslinker) for item in data]  # pyright: ignore[reportReturnType]
    _ok = check_input(data, "data", dict)
    if "data_type" not in data or data["data_type"] not in [
        "crosslink",
        "crosslink-spectrum-match",
    ]:
        raise TypeError(
            "Unsupported data type for input data! Parameter data has to be a (list of) crosslink or crosslink-spectrum-match!"
        )
    if data["data_type"] == "crosslink":
        return __to_proforma_xl(data, crosslinker)
    return __to_proforma_csm(data, crosslinker)



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings
from tqdm import tqdm
from pyteomics import mzid

from ..data import check_input
from ..data import create_csm
from ..data import create_parser_result
from ..constants import CROSSLINKERS
from .util import format_sequence

from typing import Optional
from typing import BinaryIO
from typing import Dict
from typing import Any
from typing import List
from typing import Callable

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def parse_scan_nr_from_mzid(spectrum_id: str) -> int:
    r"""Parse the scan number from a 'spectrumID' of a mzIdentML file.

    Parameters
    ----------
    title : str
        The 'spectrumID' of the mass spectrum from an mzIdentML file read with ``pyteomics``.

    Returns
    -------
    int
        The scan number.

    Examples
    --------
    >>> from pyXLMS.parser import parse_scan_nr_from_mzid
    >>> parse_scan_nr_from_mzid("scan=5321")
    5321
    """
    return int(str(spectrum_id).split("scan=")[1].split(",")[0])


def read_mzid(
    files: str | List[str] | BinaryIO,
    scan_nr_parser: Optional[Callable[[str], int]] = None,
    decoy: Optional[bool] = None,
    crosslinkers: Dict[str, float] = CROSSLINKERS,
    verbose: Literal[0, 1, 2] = 1,
) -> Dict[str, Any]:
    r"""Read a mzIdentML (mzid) file.

    Reads crosslink-spectrum-matches from a mzIdentML (mzid) file and
    returns a ``parser_result``.

    Parameters
    ----------
    files : str, list of str, or file stream
        The name/path of the mzIdentML (mzid) file(s) or a file-like object/stream.
    scan_nr_parser : callable, or None, default = None
        A function that parses the scan number from mzid spectrumIDs. If None (default)
        the function ``parse_scan_nr_from_mzid()`` is used.
    decoy : bool, or None, default = None
        Whether the mzid file contains decoy CSMs (``True``) or target CSMs (``False``).
    crosslinkers: dict of str, float, default = ``constants.CROSSLINKERS``
        Mapping of crosslinker names to crosslinker delta masses.
    verbose : 0, 1, or 2, default = 1
        - 0: All warnings are ignored.
        - 1: Warnings are printed to stdout.
        - 2: Warnings are treated as errors.

    Returns
    -------
    dict
        The ``parser_result`` object containing all parsed information.

    Raises
    ------
    RuntimeError
        If the file(s) could not be read or if the file(s) contain no crosslink-spectrum-matches.
    RuntimeError
        If parser is used with ``verbose = 2``.
    RuntimeError
        If there are warnings while reading the mzIdentML file (only for ``verbose = 2``).
    TypeError
        If parameter verbose was not set correctly.
    TypeError
        If one of the values necessary to create a crosslink-spectrum-match could not be parsed
        correctly.

    Notes
    -----
    This parser is experimental, as I don't know if the mzIdentML structure is consistent accross different
    crosslink search engines. This parser was tested with mzIdentML files from MS Annika and XlinkX.

    Warnings
    --------
    This parser only parses minimal data because most information is not available from the mzIdentML file.
    The available data is:

    - ``alpha_peptide``
    - ``alpha_peptide_crosslink_position``
    - ``beta_peptide``
    - ``beta_peptide_crosslink_position``
    - ``spectrum_file``
    - ``scan_nr``

    Examples
    --------
    >>> from pyXLMS.parser import read_mzid
    >>> csms = read_mzid("data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1.mzid")
    """
    ## check input
    _ok = (
        check_input(scan_nr_parser, "scan_nr_parser", Callable)
        if scan_nr_parser is not None
        else True
    )
    _ok = check_input(decoy, "decoy", bool) if decoy is not None else True
    _ok = check_input(crosslinkers, "crosslinkers", dict, float)
    _ok = check_input(verbose, "verbose", int)
    if verbose not in [0, 1, 2]:
        raise TypeError("Verbose level has to be one of 0, 1, or 2!")

    ## set default parsers
    if scan_nr_parser is None:
        scan_nr_parser = parse_scan_nr_from_mzid

    ## warning message
    if verbose == 1:
        warnings.warn(
            UserWarning(
                "Please be aware that mzIdentML parsing is currently an experimental feature!\n"
                "Please check the documentation for parser.read_mzid for more information!"
            )
        )
    if verbose == 2:
        raise RuntimeError(
            "Please be aware that mzIdentML parsing is currently an experimental feature!\n"
            "Please check the documentation for parser.read_mzid for more information!"
        )

    ## helper functions
    def check_str(value: str | None) -> str:
        if value is None:
            raise TypeError("Expected str value but None was given!")
        if type(value) is str:
            return value
        raise TypeError(f"Expected str value but {type(value)} was given!")
        return "err"

    def check_int(value: int | None) -> int:
        if value is None:
            raise TypeError("Expected int value but None was given!")
        if type(value) is int:
            return value
        raise TypeError(f"Expected int value but {type(value)} was given!")
        return -1

    ## data structures
    csms = list()

    ## handle input
    if not isinstance(files, list):
        inputs = [files]
    else:
        inputs = files

    ## process data
    for input in inputs:
        # read all items with pyteomics
        with warnings.catch_warnings(record=True) as wl:
            warnings.simplefilter("always")
            pyteomics_mzid = mzid.MzIdentML(input)
            items = [item for item in pyteomics_mzid]
            pyteomics_mzid.close()
        if verbose > 0 and len(wl) > 0:
            for w in wl:
                warnings.warn(w.message)
        if verbose == 2 and len(wl) > 0:
            raise RuntimeError("Reading mzIdentML file raised warnings!")
        # iterate over all items
        for item in tqdm(items):
            # set up empty variables that are needed for a minimal CSM
            csm_id = None
            scan = None
            filename = None
            peptide_a = None
            pos_a = None
            peptide_b = None
            pos_b = None
            # set scan
            if "spectrumID" in item:
                scan = scan_nr_parser(item["spectrumID"])
            # set spectrum file name
            if "location" in item:
                filename = str(item["location"]).strip()
            # check if any identifications for the spectrum
            if "SpectrumIdentificationItem" in item:
                for subitem in item["SpectrumIdentificationItem"]:
                    # we only consider rank 1 CSMs
                    if "rank" in subitem:
                        if int(subitem["rank"]) > 1:
                            continue
                    # check if item is a CSM
                    if "cross-link spectrum identification item" in subitem:
                        # if csm_id is not set yet, we parse item as alpha peptide
                        if csm_id is None:
                            csm_id = int(
                                float(
                                    subitem["cross-link spectrum identification item"]
                                )
                            )
                            if "PeptideSequence" in subitem:
                                peptide_a = format_sequence(subitem["PeptideSequence"])
                            # we only parse crosslink position from modifications
                            if "Modification" in subitem:
                                for mod in subitem["Modification"]:
                                    if "name" in mod:
                                        if (
                                            str(mod["name"]).strip().upper()
                                            in crosslinkers
                                        ):
                                            if "location" in mod:
                                                pos_a = int(mod["location"])
                        # if csm_id is already set, we check if csm_ids of items are equal,
                        # if yes we parse the item as the beta peptide
                        elif csm_id == int(
                            float(subitem["cross-link spectrum identification item"])
                        ):
                            if "PeptideSequence" in subitem:
                                peptide_b = format_sequence(subitem["PeptideSequence"])
                            if "Modification" in subitem:
                                for mod in subitem["Modification"]:
                                    if "name" in mod:
                                        if (
                                            str(mod["name"]).strip().upper()
                                            in crosslinkers
                                        ):
                                            if "location" in mod:
                                                pos_b = int(mod["location"])
            # if and only if all minimal CSM values are parsed, we create a CSM
            if None not in [csm_id, scan, filename, peptide_a, pos_a, peptide_b, pos_b]:
                csm = create_csm(
                    peptide_a=check_str(peptide_a),
                    modifications_a=None,
                    xl_position_peptide_a=check_int(pos_a),
                    proteins_a=None,
                    xl_position_proteins_a=None,
                    pep_position_proteins_a=None,
                    score_a=None,
                    decoy_a=decoy,
                    peptide_b=check_str(peptide_b),
                    modifications_b=None,
                    xl_position_peptide_b=check_int(pos_b),
                    proteins_b=None,
                    xl_position_proteins_b=None,
                    pep_position_proteins_b=None,
                    score_b=None,
                    decoy_b=decoy,
                    score=None,
                    spectrum_file=check_str(filename),
                    scan_nr=check_int(scan),
                    charge=None,
                    rt=None,
                    im_cv=None,
                )
                csms.append(csm)
    ## check results
    if len(csms) == 0:
        raise RuntimeError(
            "No crosslink-spectrum-matches were parsed! If this is unexpected, please file a bug report!"
        )
    ## return parser result
    return create_parser_result(
        search_engine="mzIdentML",
        csms=csms,
        crosslinks=None,
    )



#!/usr/bin/env python3

# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

from __future__ import annotations

import warnings
import numpy as np
from tqdm import tqdm

from ..data import check_input
from ..data import check_input_multi
from ..data import create_parser_result
from .util import get_available_keys
from .filter import filter_target_decoy
from .filter import filter_crosslink_type

from typing import Dict
from typing import List
from typing import Any

# legacy
try:
    from typing import Literal
except ImportError:
    from typing_extensions import Literal


def __verify_fdr_strict(
    data: List[Dict[str, Any]],
    fdr: float,
    cutoff: float,
    score: Literal["higher_better", "lower_better"],
) -> bool:
    r"""Verifies that a list of crosslinks or crosslink-spectrum-matches has the estimated false discovery rate using a strict approach.

    Verifies that a list of crosslinks or crosslink-spectrum-matches has the estimated false discovery rate (FDR) using the
    formula (TD+DD)/TT given a score cutoff.
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to validate.
    fdr : float
        The target FDR, must be given as a real number between 0 and 1.
    cutoff : float
        Score cutoff that defines which crosslinks or crosslink-spectrum-matches fall within the FDR validated result.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    bool
        Returns True if the given score cutoff yields only crosslinks or crosslink-spectrum-matches
        that are within the target FDR. Returns False if the given score cutoff yields crosslinks or
        crosslink-spectrum-matches that produce a higher estimated FDR than the desired target FDR.

    Notes
    -----
    This function should not be called directly, it is called from ``__validate_strict()``.
    """
    D = 0
    T = 0
    for item in data:
        if score == "higher_better" and item["score"] >= cutoff:
            if not item["alpha_decoy"] and not item["beta_decoy"]:
                T += 1
            else:
                D += 1
        elif score == "lower_better" and item["score"] <= cutoff:
            if not item["alpha_decoy"] and not item["beta_decoy"]:
                T += 1
            else:
                D += 1
        else:
            # do nothing
            pass
    return D / T < fdr


def __validate_strict(
    data: List[Dict[str, Any]],
    fdr: float,
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Validate a list of crosslinks or crosslink-spectrum-matches by strict false discovery rate estimation.

    Validate a list of crosslinks or crosslink-spectrum-matches by strict false discovery rate (FDR) estimation using the
    formula (TD+DD)/TT.
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to validate.
    fdr : float
        The target FDR, must be given as a real number between 0 and 1. The default of 0.01 corresponds to 1% FDR.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        A list of validated crosslink-spectrum-matches or crosslinks.

    Warns
    -----
    RuntimeWarning
        If none of the data passes the desired FDR threshold.

    Notes
    -----
    This function should not be called directly, it is called from ``validate()``.
    """
    scores = list()
    td = list()
    for item in data:
        scores.append(item["score"])
        if not item["alpha_decoy"] and not item["beta_decoy"]:
            td.append(0)
        else:
            td.append(1)
    scores = np.array(scores)
    cutoff = 0
    if score == "higher_better":
        td = np.array(td)[np.argsort(scores, kind="stable")]
        scores = scores[np.argsort(scores, kind="stable")]
        cutoff = scores[0]  # scores.max()
    else:
        td = np.array(td)[np.argsort(scores, kind="stable")[::-1]]
        scores = scores[np.argsort(scores, kind="stable")[::-1]]
        cutoff = scores[0]  # scores.min()
    nr_items = len(td)
    for i in tqdm(
        range(nr_items),
        total=nr_items,
        desc="Iterating over scores for FDR calculation...",
    ):
        if (nr_items - i - td[i:].sum()) <= 0.0:
            warnings.warn(
                RuntimeWarning(
                    "None of the data passes the desired FDR threshold! This is usually due to decoys with very good scores."
                )
            )
            return []
        if td[i:].sum() / (nr_items - i - td[i:].sum()) < fdr:
            # we need to verify in this case because there might be multiple
            # items with the same score
            if __verify_fdr_strict(data, fdr, scores[i], score):
                cutoff = scores[i]
                break
    validated_items = list()
    for item in data:
        if score == "higher_better" and item["score"] >= cutoff:
            validated_items.append(item)
        elif score == "lower_better" and item["score"] <= cutoff:
            validated_items.append(item)
        else:
            # do nothing
            pass
    return validated_items


def __verify_fdr_relaxed(
    data: List[Dict[str, Any]],
    fdr: float,
    cutoff: float,
    score: Literal["higher_better", "lower_better"],
) -> bool:
    r"""Verifies that a list of crosslinks or crosslink-spectrum-matches has the estimated false discovery rate using a relaxed approach.

    Verifies that a list of crosslinks or crosslink-spectrum-matches has the estimated false discovery rate (FDR) using the
    formula (TD-DD)/TT given a score cutoff.
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to validate.
    fdr : float
        The target FDR, must be given as a real number between 0 and 1.
    cutoff : float
        Score cutoff that defines which crosslinks or crosslink-spectrum-matches fall within the FDR validated result.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    bool
        Returns True if the given score cutoff yields only crosslinks or crosslink-spectrum-matches
        that are within the target FDR. Returns False if the given score cutoff yields crosslinks or
        crosslink-spectrum-matches that produce a higher estimated FDR than the desired target FDR.

    Raises
    ------
    RuntimeError
        If the number of DD matches exceeds the number of TD matches.
        FDR can not be estimated with the formula '(TD-DD)/TT' in these cases.

    Notes
    -----
    This function should not be called directly, it is called from ``__validate_relaxed()``.
    """
    D = 0
    DT = 0
    T = 0
    for item in data:
        if score == "higher_better" and item["score"] >= cutoff:
            if not item["alpha_decoy"] and not item["beta_decoy"]:
                T += 1
            elif item["alpha_decoy"] and item["beta_decoy"]:
                D += 1
            else:
                DT += 1
        elif score == "lower_better" and item["score"] <= cutoff:
            if not item["alpha_decoy"] and not item["beta_decoy"]:
                T += 1
            elif item["alpha_decoy"] and item["beta_decoy"]:
                D += 1
            else:
                DT += 1
        else:
            # do nothing
            pass
    if (DT - D) < 0.0:
        raise RuntimeError(
            f"Number of DD matches is greater than the number of TD matches for score {cutoff}! "
            "Invalid FDR estimation! Please use formula 'D/T' instead!"
        )
    return (DT - D) / T < fdr


def __validate_relaxed(
    data: List[Dict[str, Any]],
    fdr: float,
    score: Literal["higher_better", "lower_better"],
) -> List[Dict[str, Any]]:
    r"""Validate a list of crosslinks or crosslink-spectrum-matches by relaxed false discovery rate estimation.

    Validate a list of crosslinks or crosslink-spectrum-matches by relaxed false discovery rate (FDR) estimation using the
    formula (TD-DD)/TT.
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to validate.
    fdr : float
        The target FDR, must be given as a real number between 0 and 1.
    score : str, one of "higher_better" or "lower_better"
        If a higher score is considered better, or a lower score is considered better.

    Returns
    -------
    list of dict of str, any
        A list of validated crosslink-spectrum-matches or crosslinks.

    Raises
    ------
    RuntimeError
        If the number of DD matches exceeds the number of TD matches.
        FDR can not be estimated with the formula '(TD-DD)/TT' in these cases.

    Warns
    -----
    RuntimeWarning
        If none of the data passes the desired FDR threshold.

    Notes
    -----
    This function should not be called directly, it is called from ``validate()``.
    """
    scores = list()
    td = list()
    tdd = list()
    for item in data:
        scores.append(item["score"])
        if not item["alpha_decoy"] and not item["beta_decoy"]:
            td.append(0)
            tdd.append(0)
        elif item["alpha_decoy"] and item["beta_decoy"]:
            td.append(1)
            tdd.append(-1)
        else:
            td.append(1)
            tdd.append(1)
    scores = np.array(scores)
    cutoff = 0
    if score == "higher_better":
        td = np.array(td)[np.argsort(scores, kind="stable")]
        tdd = np.array(tdd)[np.argsort(scores, kind="stable")]
        scores = scores[np.argsort(scores, kind="stable")]
        cutoff = scores[0]  # scores.max()
    else:
        td = np.array(td)[np.argsort(scores, kind="stable")[::-1]]
        tdd = np.array(tdd)[np.argsort(scores, kind="stable")[::-1]]
        scores = scores[np.argsort(scores, kind="stable")[::-1]]
        cutoff = scores[0]  # scores.min()
    nr_items = len(td)
    for i in tqdm(
        range(nr_items),
        total=nr_items,
        desc="Iterating over scores for FDR calculation...",
    ):
        if tdd[i:].sum() < 0.0:
            raise RuntimeError(
                f"Number of DD matches is greater than the number of TD matches for score {scores[i]}! "
                "Invalid FDR estimation! Please use formula 'D/T' instead!"
            )
        if (nr_items - i - td[i:].sum()) <= 0.0:
            warnings.warn(
                RuntimeWarning(
                    "None of the data passes the desired FDR threshold! This is usually due to decoys with very good scores."
                )
            )
            return []
        if tdd[i:].sum() / (nr_items - i - td[i:].sum()) < fdr:
            # we need to verify in this case because there might be multiple
            # items with the same score
            if __verify_fdr_relaxed(data, fdr, scores[i], score):
                cutoff = scores[i]
                break
    validated_items = list()
    for item in data:
        if score == "higher_better" and item["score"] >= cutoff:
            validated_items.append(item)
        elif score == "lower_better" and item["score"] <= cutoff:
            validated_items.append(item)
        else:
            # do nothing
            pass
    return validated_items


def validate(
    data: List[Dict[str, Any]] | Dict[str, Any],
    fdr: float = 0.01,
    formula: Literal["D/T", "(TD+DD)/TT", "(TD-DD)/TT"] = "D/T",
    score: Literal["higher_better", "lower_better"] = "higher_better",
    separate_intra_inter: bool = False,
    ignore_missing_labels: bool = False,
) -> List[Dict[str, Any]] | Dict[str, Any]:
    r"""Validate a list of crosslinks or crosslink-spectrum-matches, or a parser_result by estimating false discovery rate.

    Validate a list of crosslinks or crosslink-spectrum-matches, or a parser_result by estimating false discovery rate (FDR) using the defined formula.
    Requires that "score", "alpha_decoy" and "beta_decoy" fields are set for crosslinks and crosslink-spectrum-matches.

    Parameters
    ----------
    data : list of dict of str, any, or dict of str, any
        A list of crosslink-spectrum-matches or crosslinks to validate, or a parser_result.
    fdr : float, default = 0.01
        The target FDR, must be given as a real number between 0 and 1. The default of 0.01 corresponds to 1% FDR.
    formula : str, one of "D/T", "(TD+DD)/TT", or "(TD-DD)/TT", default = "D/T"
        Which formula to use to estimate FDR. D and DD denote decoy matches, T and TT denote target matches, and TD denotes target-decoy
        and decoy-target matches.
    score : str, one of "higher_better" or "lower_better", default = "higher_better"
        If a higher score is considered better, or a lower score is considered better.
    separate_intra_inter : bool, default = False
        If FDR should be estimated separately for intra and inter matches.
    ignore_missing_labels : bool, default = False
        If crosslinks and crosslink-spectrum-matches should be ignored if they don't have target and decoy labels. By default and error is
        thrown if any unlabelled data is encountered.

    Returns
    -------
    list of dict of str, any, or dict of str, any
        If a list of crosslink-spectrum-matches or crosslinks was provided, a list of validated
        crosslink-spectrum-matches or crosslinks is returned. If a parser_result was provided,
        an parser_result with validated crosslink-spectrum-matches and/or validated crosslinks will
        be returned.

    Raises
    ------
    TypeError
        If a wrong data type is provided.
    TypeError
        If parameter formula is not one of 'D/T', '(TD+DD)/TT', or '(TD-DD)/TT'.
    TypeError
        If parameter score is not one of 'higher_better' or 'lower_better'.
    ValueError
        If parameter fdr is outside of the supported range.
    ValueError
        If attribute 'score' is not available for any of the data.
    ValueError
        If attribute 'alpha_decoy' or 'beta_decoy' is not available for any of the data and parameter ignore_missing_labels
        is set to False.
    ValueError
        If the number of DD matches exceeds the number of TD matches for formula '(TD-DD)/TT'.
        FDR cannot be estimated with the formula '(TD-DD)/TT' in these cases.

    Notes
    -----
    Please note that progress bars will usually not complete when running this function. This is by design as it is not
    necessary to iterate over all scores to estimate FDR.

    Examples
    --------
    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import validate
    >>> pr = read(
    ...     "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> csms = pr["crosslink-spectrum-matches"]
    >>> len(csms)
    826
    >>> validated = validate(csms)
    >>> len(validated)
    705

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import validate
    >>> pr = read(
    ...     [
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     ],
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> len(pr["crosslink-spectrum-matches"])
    826
    >>> len(pr["crosslinks"])
    300
    >>> validated = validate(pr)
    >>> len(validated["crosslink-spectrum-matches"])
    705
    >>> len(validated["crosslinks"])
    226

    >>> from pyXLMS.parser import read
    >>> from pyXLMS.transform import validate
    >>> pr = read(
    ...     [
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
    ...         "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
    ...     ],
    ...     engine="MS Annika",
    ...     crosslinker="DSS",
    ... )
    >>> len(pr["crosslink-spectrum-matches"])
    826
    >>> len(pr["crosslinks"])
    300
    >>> validated = validate(pr, fdr=0.05)
    >>> len(validated["crosslink-spectrum-matches"])
    825
    >>> len(validated["crosslinks"])
    260
    """
    _ok = check_input_multi(data, "data", [dict, list])
    _ok = check_input(fdr, "fdr", float)
    _ok = check_input(formula, "formula", str)
    _ok = check_input(score, "score", str)
    _ok = check_input(separate_intra_inter, "separate_intra_inter", bool)
    _ok = check_input(ignore_missing_labels, "ignore_missing_labels", bool)
    if fdr >= 1.0 or fdr <= 0.0:
        raise ValueError(
            "FDR must be given as a real number between 0 and 1, e.g. 0.01 corresponds to 1% FDR!"
        )
    if formula not in ["D/T", "(TD+DD)/TT", "(TD-DD)/TT"]:
        raise TypeError(
            "Parameter 'formula' has to be one of 'D/T', '(TD+DD)/TT' or '(TD-DD)/TT'! Where D and DD is the number of decoys, T and TT the number of targets, "
            "and TD the number of target-decoys!"
        )
    if score not in ["higher_better", "lower_better"]:
        raise TypeError(
            "Parameter 'score' has to be one of 'higher_better' or 'lower_better'!"
        )
    if isinstance(data, list):
        _ok = check_input(data, "data", list, dict)
        if len(data) == 0:
            return data
        if "data_type" not in data[0] or data[0]["data_type"] not in [
            "crosslink",
            "crosslink-spectrum-match",
        ]:
            raise TypeError(
                "Unsupported data type for input data! Parameter data has to be a list of crosslink or crosslink-spectrum-match, "
                "or a parser_result!"
            )
        if ignore_missing_labels:
            data = [
                item
                for item in data
                if item["alpha_decoy"] is not None and item["beta_decoy"] is not None
            ]
        available_keys = get_available_keys(data)
        if (
            not available_keys["score"]
            or not available_keys["alpha_decoy"]
            or not available_keys["beta_decoy"]
        ):
            raise ValueError(
                "Can't validate data if 'score' or target/decoy labels are missing! Selecting 'ignore_missing_labels = True' will ignore crosslinks and crosslink-spectrum-matches "
                "that don't have a valid target/decoy label and filter them out!"
            )
        if formula == "(TD-DD)/TT":
            if len(filter_target_decoy(data)["Target-Decoy"]) == 0:
                raise ValueError(
                    "Can't estimate FDR with formula '(TD-DD)/TT' when there are no TD matches! Please select the default formula instead!"
                )
            if separate_intra_inter:
                separate = filter_crosslink_type(data)
                return __validate_relaxed(
                    separate["Intra"], fdr, score
                ) + __validate_relaxed(separate["Inter"], fdr, score)
            return __validate_relaxed(data, fdr, score)
        if separate_intra_inter:
            separate = filter_crosslink_type(data)
            return __validate_strict(separate["Intra"], fdr, score) + __validate_strict(
                separate["Inter"], fdr, score
            )
        return __validate_strict(data, fdr, score)
    if "data_type" not in data or data["data_type"] != "parser_result":
        raise TypeError("Can't validate dict. Dict has to be a valid 'parser_result'!")
    new_csms = (
        validate(
            data["crosslink-spectrum-matches"],
            fdr,
            formula,
            score,
            separate_intra_inter,
            ignore_missing_labels,
        )
        if data["crosslink-spectrum-matches"] is not None
        else None
    )
    new_xls = (
        validate(
            data["crosslinks"],
            fdr,
            formula,
            score,
            separate_intra_inter,
            ignore_missing_labels,
        )
        if data["crosslinks"] is not None
        else None
    )
    if new_csms is not None:
        if not isinstance(new_csms, list):
            raise RuntimeError(
                "Something went wrong while validating crosslink-spectrum-matches.\n"
                f"Expected data type: list. Got: {type(new_csms)}."
            )
    if new_xls is not None:
        if not isinstance(new_xls, list):
            raise RuntimeError(
                "Something went wrong while validating crosslinks.\n"
                f"Expected data type: list. Got: {type(new_xls)}."
            )
    return create_parser_result(
        search_engine=data["search_engine"], csms=new_csms, crosslinks=new_xls
    )



#!/usr/bin/env python3

# pyXLMS - TESTS
# 2025 (c) Micha Johannes Birklbauer
# https://github.com/michabirklbauer/
# micha.birklbauer@gmail.com

import pytest


def test1():
    from pyXLMS.parser import read
    from pyXLMS.transform import summary

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    csms = pr["crosslink-spectrum-matches"]
    stats = summary(csms)
    assert stats["Number of CSMs"] == pytest.approx(826.0)
    assert stats["Number of unique CSMs"] == pytest.approx(826.0)
    assert stats["Number of intra CSMs"] == pytest.approx(803.0)
    assert stats["Number of inter CSMs"] == pytest.approx(23.0)
    assert stats["Number of target-target CSMs"] == pytest.approx(786.0)
    assert stats["Number of target-decoy CSMs"] == pytest.approx(39.0)
    assert stats["Number of decoy-decoy CSMs"] == pytest.approx(1.0)
    assert stats["Minimum CSM score"] == pytest.approx(1.11)
    assert stats["Maximum CSM score"] == pytest.approx(452.99)


def test2():
    from pyXLMS.parser import read
    from pyXLMS.transform import summary

    pr = read(
        "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        engine="MS Annika",
        crosslinker="DSS",
    )
    stats = summary(pr)
    assert stats["Number of crosslinks"] == pytest.approx(300.0)
    assert stats["Number of unique crosslinks by peptide"] == pytest.approx(300.0)
    assert stats["Number of unique crosslinks by protein"] == pytest.approx(298.0)
    assert stats["Number of intra crosslinks"] == pytest.approx(279.0)
    assert stats["Number of inter crosslinks"] == pytest.approx(21.0)
    assert stats["Number of target-target crosslinks"] == pytest.approx(265.0)
    assert stats["Number of target-decoy crosslinks"] == pytest.approx(0.0)
    assert stats["Number of decoy-decoy crosslinks"] == pytest.approx(35.0)
    assert stats["Minimum crosslink score"] == pytest.approx(1.11)
    assert stats["Maximum crosslink score"] == pytest.approx(452.99)


def test3():
    from pyXLMS.parser import read
    from pyXLMS.transform import summary

    pr = read(
        [
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_CSMs.xlsx",
            "data/ms_annika/XLpeplib_Beveridge_QEx-HFX_DSS_R1_Crosslinks.xlsx",
        ],
        engine="MS Annika",
        crosslinker="DSS",
    )
    stats = summary(pr)
    assert stats["Number of CSMs"] == pytest.approx(826.0)
    assert stats["Number of unique CSMs"] == pytest.approx(826.0)
    assert stats["Number of intra CSMs"] == pytest.approx(803.0)
    assert stats["Number of inter CSMs"] == pytest.approx(23.0)
    assert stats["Number of target-target CSMs"] == pytest.approx(786.0)
    assert stats["Number of target-decoy CSMs"] == pytest.approx(39.0)
    assert stats["Number of decoy-decoy CSMs"] == pytest.approx(1.0)
    assert stats["Minimum CSM score"] == pytest.approx(1.11)
    assert stats["Maximum CSM score"] == pytest.approx(452.99)
    assert stats["Number of crosslinks"] == pytest.approx(300.0)
    assert stats["Number of unique crosslinks by peptide"] == pytest.approx(300.0)
    assert stats["Number of unique crosslinks by protein"] == pytest.approx(298.0)
    assert stats["Number of intra crosslinks"] == pytest.approx(279.0)
    assert stats["Number of inter crosslinks"] == pytest.approx(21.0)
    assert stats["Number of target-target crosslinks"] == pytest.approx(265.0)
    assert stats["Number of target-decoy crosslinks"] == pytest.approx(0.0)
    assert stats["Number of decoy-decoy crosslinks"] == pytest.approx(35.0)
    assert stats["Minimum crosslink score"] == pytest.approx(1.11)
    assert stats["Maximum crosslink score"] == pytest.approx(452.99)
